/*************************************************************************
 *                                                                       *
 * GLE - Graphics Layout Engine <http://www.gle-graphics.org/>           *
 *                                                                       *
 * BSD License                                                           *
 *                                                                       *
 * Copyright (c).                                                        *
 * All rights reserved.                                                  *
 *                                                                       *
 * Redistribution and use in source and binary forms, with or without    *
 * modification, are permitted provided that the following conditions    *
 * are met:                                                              *
 *                                                                       *
 *  * Redistributions of source code must retain the above copyright     *
 *  notice, this list of conditions and the following disclaimer.        *
 *                                                                       *
 *  * Redistributions in binary form must reproduce the above            *
 *  copyright notice, this list of conditions and the following          *
 *  disclaimer in the documentation and/or other materials provided      *
 *  with the distribution.                                               *
 *                                                                       *
 *  * Neither the name of GLE nor the names of its contributors may      *
 *  be used to endorse or promote products derived from this software    *
 *  without specific prior written permission.                           *
 *                                                                       *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR *
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT  *
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, *
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT      *
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, *
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY *
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT   *
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE *
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.  *
 *                                                                       *
 *************************************************************************/

#include "all.h"
#include "tokens/stokenizer.h"
#include "SourceLine.h"
#include "justify.h"
#include "mygraph.h"
#include "core.h"
#include "axis.h"
#include "gprint.h"
#include "d_interface.h"
#include "numberformat.h"
#include "cutils.h"

#define GRAPHDEF extern
#include "mem_limits.h"
#include "token.h"
#include "graph.h"

void start_subtick(double *tick1, double gmin, double dticks);
void nice_ticks(double *dticks, double *gmin,double *gmax, double *t1,double *tn,int minset, int maxset);
void numtrim(char **d,char *s, double dticks);
void nice_log_ticks(double *start, double *last, double gmin, double gmax) throw (ParserError);

double fnloglen(double v,axis_struct *ax);
double fnlogx(double v,axis_struct *ax);
int fnbig(double v);

void draw_axis_titles(axis_struct *ax, double h, double ox, double oy, GLEMeasureBox measure);
void draw_axis_titles_v35(axis_struct *ax, double h, double ox, double oy, double dticks, double llen);

extern double graph_x1;
extern double graph_x2;
extern double graph_y1;
extern double graph_y2;

#define fnx(vv) m_fnx(vv)
#define fnlx(vv) fnlogx(vv,ax)
#define true (!false)
#define false 0

static double  gglen,ggmin,ggmax,ggnegate;

double m_fnx(double v);

double m_fnx(double v) {
	if (ggnegate) 	{
		v = ggmax - (v-ggmin);
	}
	return (((v-ggmin)/(ggmax-ggmin)) * gglen );
}

bool axis_is_pos(double pos, int* cnt, double del, vector<double>& vec) {
	if (*cnt < (int)vec.size()) {
		while (*cnt < (int)vec.size() && pos > vec[*cnt]+del/100.0) {
			(*cnt)++;
		}
		if (*cnt < (int)vec.size() && fabs(pos-vec[*cnt]) < del/100.0) {
			return true;
		}
	}
	return false;
}

bool axis_is_pos_perc(double pos, int* cnt, double perc, vector<double>& vec) {
	if (*cnt < (int)vec.size()) {
		while (*cnt < (int)vec.size() && pos > vec[*cnt]*(1+perc)) {
			(*cnt)++;
		}
		if (*cnt < (int)vec.size()) {
			if (vec[*cnt] != 0.0) {
				if (fabs((pos-vec[*cnt])/vec[*cnt]) < perc) return true;
			} else {
				if (fabs(pos-vec[*cnt]) < perc) return true;
			}
		}
	}
	return false;
}

bool equals_rel(double v1, double v2) {
	if (v1 != 0.0) {
		return fabs(v1 - v2)/v1 < 1e-6;
	} else {
		return fabs(v1 - v2) < 1e-6;
	}
}

void axis_draw_tick(axis_struct *ax, double fi, int* tick1_cnt, int* tick2_cnt, double ox, double oy, double t) {
	bool has_tick1 = t > 0.0 || ax->ticks_both;
	bool has_tick2 = t < 0.0 || ax->ticks_both;
	has_tick1 &= !ax->isNoTick1(fi, tick1_cnt, ax->dsubticks);
	has_tick2 &= !ax->isNoTick2(fi, tick2_cnt, ax->dsubticks);
	if (!has_tick1 && !has_tick2) return;
	double inv = axis_ticks_neg(ax->type) ? -1.0 : 1.0;
	double from = has_tick2 ? -fabs(t)*inv : 0.0;
	double to = has_tick1 ? fabs(t)*inv : 0.0;
	if (axis_horizontal(ax->type)) {
		g_move(ox+fnx(fi), oy+from);
		g_line(ox+fnx(fi), oy+to);
	} else {
		g_move(ox+from, oy+fnx(fi));
		g_line(ox+to, oy+fnx(fi));
	}
}

void axis_draw_tick_log(axis_struct *ax, double fi, int* tick1_cnt, int* tick2_cnt, double ox, double oy, double t) {
	bool has_tick1 = t > 0.0 || ax->ticks_both;
	bool has_tick2 = t < 0.0 || ax->ticks_both;
	has_tick1 &= !ax->isNoTick1Perc(fi, tick1_cnt);
	has_tick2 &= !ax->isNoTick2Perc(fi, tick2_cnt);
	if (!has_tick1 && !has_tick2) return;
	double inv = axis_ticks_neg(ax->type) ? -1.0 : 1.0;
	double from = has_tick2 ? -fabs(t)*inv : 0.0;
	double to = has_tick1 ? fabs(t)*inv : 0.0;
	if (axis_horizontal(ax->type)) {
		g_move(ox+fnlx(fi), oy+from);
		g_line(ox+fnlx(fi), oy+to);
	} else {
		g_move(ox+from, oy+fnlx(fi));
		g_line(ox+to, oy+fnlx(fi));
	}
}

void draw_axis(axis_struct *ax, gbox* box, bool drawgrid) {
	double fi,x,y,gmin,gmax,dticks,tick1,tickn;
	int i,xax,n,savecap;
	double tlen,stlen,dsubticks,tt,t;
	double h,dist;
	double bl,br,bu,bd;
	double ox,oy;

	// Fix for font problems by Jan Struyf
	g_resetfont();

	xax = axis_horizontal(ax->type);
	g_source("BEGIN AXIS  =====\n");
	if (ax->off) return;
	g_get_xy(&ox,&oy);

	bool drawticks = true;
	bool drawrest = true;

	if (ax->hasGrid()) {
		if (drawgrid) drawrest = false;
		else drawticks = false;
	} else {
		if (drawgrid) return;
	}

/*----------------------------- Generate the places for labels to go */
	ggmin = ax->min;
	ggmax = ax->max;
	ggnegate = ax->negate;
	gmin = ax->min;
	gmax = ax->max;
	gglen = ax->length;
	int lgset = ax->lgset;

	if (ax->log) {
		/* generate places for log main ticks */
		if (lgset == GLE_AXIS_LOG_DEFAULT) {
			/* dist between 10^0 and 10^1 */
			dist = fnloglen(2,ax) - fnloglen(1,ax);
			if (dist > ax->base*4.5) lgset = GLE_AXIS_LOG_25;
			if (dist > ax->base*22.0) lgset = GLE_AXIS_LOG_1;
		}
		h = ax->label_hei;
		if (h==0) h = ax->base;
		nice_log_ticks(&tick1, &tickn, gmin, gmax);
		if (lgset == GLE_AXIS_LOG_DEFAULT || lgset == GLE_AXIS_LOG_OFF) {
			dticks = (tickn - tick1) / gglen * 3 * h;
			if (dticks >= 4) {
				dticks = ((int)dticks / 5) * 5.0;
			} else if (dticks >= 2) {
				dticks = 2;
			}
		} else {
			dticks = 1.0;
		}
		if (ax->nticks != 0) {
			dticks = (tickn - tick1)/(ax->nticks - 1);
		}
		if (ax->dticks != 0) {
			dticks = ax->dticks;
		}
		if (dticks < 1.0) {
			/* values smaller than 0.9 translate to lgset */
			if (dticks < 0.2) lgset = GLE_AXIS_LOG_1;
			else lgset = GLE_AXIS_LOG_25;
			dticks = 1.0;
		}
		if (dticks > 1.0 || ax->subticks_off) {
			/* if dticks too large or subticks disabled */
			lgset = GLE_AXIS_LOG_OFF;
			dticks = floor(dticks);
		}
		if (ax->getNbPlaces() == 0) {
			for (double pw10 = tick1; pw10 <= tickn; pw10 += dticks) {
				double pos = pow(10.0, pw10);
				if (equals_rel(gmin, pos)) {
					/* first tick */
					if (!ax->nofirst) {
						ax->addPlace(pos);
					}
				} else if (equals_rel(gmax, pos)) {
					/* last tick */
					if (!ax->nolast) {
						ax->addPlace(pos);
					}
				} else {
					/* add if in window */
					if (pos >= gmin && pos <= gmax) {
						ax->addPlace(pos);
					}
				}
				/* also label subticks? */
				if (lgset == GLE_AXIS_LOG_1 || lgset == GLE_AXIS_LOG_25 || lgset == GLE_AXIS_LOG_25B) {
					for (int i = 2; i <= 9; i++) {
						if (lgset == GLE_AXIS_LOG_1 || i == 2 || i == 5) {
							double pos_sub = pos * i;
							if (pos_sub >= gmin && pos_sub <= gmax) {
								ax->addPlace(pos_sub);
							}
						}
					}
				}
			}
		}
	} else {
		dticks = 0;
		if (ax->nticks != 0) dticks = (gmax - gmin)/ax->nticks;
		if (ax->dticks != 0) dticks = ax->dticks;
		if (ax->has_ftick) {
			tick1 = ax->ftick;
			if (ax->dticks == 0) {
				// Compute suitable value for dticks
				double dummy = 0.0;
				nice_ticks(&dticks, &gmin, &gmax, &dummy, &tickn, ax->minset, ax->maxset);
			}
			if (ax->nticks != 0) {
				tickn = tick1 + (ax->nticks-1) * dticks;
			} else {
				tickn = tick1 + floor((gmax-tick1)/dticks) * dticks;
			}
		} else {
			nice_ticks(&dticks, &gmin, &gmax, &tick1, &tickn, ax->minset, ax->maxset);
		}
		if (ax->nofirst) tick1 = tick1 + dticks;
		if (ax->nolast) tickn = tickn - dticks;
		tickn = tickn + dticks/100.0;
		if (ax->getNbPlaces() == 0) {
			/* haven't been set, so generate positions. */
			double end_tick = (tickn+dticks/100.0);
			for (fi = tick1; fi <= end_tick; fi += dticks) {
				ax->addPlace(fi);
			}
		}
	}

/*------------------------------  Now draw the ticks	(subticks first) */
	g_gsave();

	// ticks length
	tlen = ax->base*g_get_fconst(GLEC_TICKSSCALE);
	if (ax->ticks_length!=0) tlen = ax->ticks_length;

	// subticks length and distance
	stlen = tlen/2;
	if (ax->subticks_length!=0) stlen = ax->subticks_length;
	dsubticks = 0;

	// draw subticks
	if (ax->log) {
		if (!ax->subticks_off && drawticks) {
			g_set_color(ax->subticks_color);
			g_set_line_width(ax->subticks_lwidth);
			g_set_line_style(ax->subticks_lstyle);
			int place_cnt = 0;
			int tick1_cnt = 0;
			int tick2_cnt = 0;
			for (double pw10 = tick1; pw10 <= tickn; pw10 += 1.0) {
				double pos = pow(10.0, pw10);
				if (pos > gmin && pos < gmax && !ax->isPlaceRel(pos, &place_cnt)) {
					axis_draw_tick_log(ax, pos, &tick1_cnt, &tick2_cnt, ox, oy, stlen);
				}
				if (lgset == GLE_AXIS_LOG_1 || lgset == GLE_AXIS_LOG_25 || lgset == GLE_AXIS_LOG_25B) {
					for (int i = 2; i <= 9; i++) {
						double pos_sub = pos * i;
						if (pos_sub > gmin && pos_sub < gmax) {
							axis_draw_tick_log(ax, pos_sub, &tick1_cnt, &tick2_cnt, ox, oy, stlen);
						}
					}
				}
			}
		}
	} else {
		if (ax->nsubticks != 0) dsubticks = dticks/(ax->nsubticks+1);
		if (ax->dsubticks != 0) dsubticks = ax->dsubticks;
		if (dsubticks == 0) dsubticks = dticks / 2.0;
		ax->dsubticks = dsubticks;
		double end_tick = gmax;
		if (ax->has_ftick) {
			tick1 += dsubticks;
			end_tick = tickn;
		} else {
			start_subtick(&tick1, gmin, dsubticks);
		}
		t = stlen;
		if (!ax->subticks_off && drawticks) {
			int place_cnt = 0;
			int tick1_cnt = 0;
			int tick2_cnt = 0;
			g_set_color(ax->subticks_color);
			g_set_line_width(ax->subticks_lwidth);
			g_set_line_style(ax->subticks_lstyle);
			for (fi = tick1; fi <= end_tick; fi += dsubticks) {
				if (!ax->isPlace(fi, &place_cnt, dsubticks)) {
					axis_draw_tick(ax, fi, &tick1_cnt, &tick2_cnt, ox, oy, t);
				}
			}
		}
	}

/*------------------------------  Now the main ticks */
	t = tlen;
	tt = stlen;
	if (!ax->ticks_off && drawticks) {
		g_set_color(ax->ticks_color);
		g_set_line_width(ax->ticks_lwidth);
		g_set_line_style(ax->ticks_lstyle);
		int tick1_cnt = 0;
		int tick2_cnt = 0;
		if (ax->log) {
			/* Draw log ticks */
			for (i = 0; i < ax->getNbPlaces(); i++) {
				fi = ax->places[i];
				if (fnbig(fi)) {
					/* only draw major ticks */
					axis_draw_tick_log(ax, fi, &tick1_cnt, &tick2_cnt, ox, oy, tlen);
				}
			}
		} else {
			for (i = 0; i < ax->getNbPlaces(); i++) {
				fi = ax->places[i];
				axis_draw_tick(ax, fi, &tick1_cnt, &tick2_cnt, ox, oy, tlen);
			}
		}
	}
	g_grestore();

/*-----------------------------  Ok lets draw the side now */
/*-----------------------------  Note: after the ticks - this gives better result if ticks are different color */
	if (!ax->side_off && drawrest) {
		g_get_line_cap(&savecap);
		g_gsave();
		g_set_color(ax->side_color);
		g_set_line_width(ax->side_lwidth);
		g_set_line_style(ax->side_lstyle);
		g_set_line_cap(1);
		g_get_xy(&x,&y);
		if (xax) g_line(x+ax->length,y);
		else g_line(x,y+ax->length);
		g_grestore();
		g_set_line_cap(savecap);
	}

/*------------------------------  Now draw the labels */
	double real_llen = 0;
	if (!ax->ticks_off) {
		if (tlen < 0) real_llen = -tlen;
		if (ax->ticks_both) real_llen = +tlen;
	}
	h = ax->label_hei;
	if (h==0) h = ax->base;
	double llen = real_llen + h*g_get_fconst(GLEC_ALABELDIST);
	if (ax->label_dist!=0) llen = ax->label_dist;
	if (ax->getNbNames() == 0) {
		char cbuff[100];
		GLENumberFormat* format = ax->format == "" ? NULL : new GLENumberFormat(ax->format);
		if (ax->log) {
			for (i = 0; i < ax->getNbPlaces(); i++) {
				fi = ax->places[i];
				n = (int) floor(.0001 + fi/pow(10.0,floor(log10(fi))));
				if (fnbig(fi)) {
					n = (int) floor(log10(fi)+0.5);
					if (format != NULL) {
						format->format(pow(10.0, n), ax->getNamePtr(i));
					} else {
						if (n == 0) {
							sprintf(cbuff,"1");
						} else if (n == 1) {
							sprintf(cbuff,"10");
						} else {
							sprintf(cbuff,"10\\sup {%d}",n);
						}
						ax->setName(i, cbuff);
					}
				} else {
					if (ax->lgset == GLE_AXIS_LOG_25B && format != NULL) {
						format->format(fi, ax->getNamePtr(i));
					} else {
						n = (int) floor(.0001 + fi/(pow(10.0,floor(log10(fi)))));
						sprintf(cbuff,"{\\sethei{%g}%d}",h*.7,n);
						ax->setName(i, cbuff);
					}
				}
			}
		} else {
			if (ax->getNamesDataSet() != -1) {
				// Get axis labels from a data set
				ax->getLabelsFromDataSet(ax->getNamesDataSet());
			} else {
				for (i = 0; i < ax->getNbPlaces(); i++) {
					fi = ax->places[i]; x = fabs(fi);
					if (x < .00001*dticks) {
						fi = 0; x = 0;
					}
					if (format != NULL) {
						format->format(fi, ax->getNamePtr(i));
					} else {
						char* num_trim_str = NULL;
						if (fi==0 || (x>1e-5 && x<1e6)) sprintf(cbuff,"%f",fi);
						else sprintf(cbuff,"%e",fi);
						numtrim(&num_trim_str,cbuff,dticks);
						if (num_trim_str != NULL) {
							ax->setName(i, num_trim_str);
							myfree(num_trim_str);
						}
					}
				}
			}
		}
		if (format != NULL) delete format;
	}
	g_set_color(ax->label_color);
	g_set_line_width(ax->label_lwidth);
	g_set_line_style(ax->label_lstyle);
	g_set_font(ax->label_font);
	g_set_hei(h*g_get_fconst(GLEC_ALABELSCALE));
	double maxd = 0, maxh = 0;
	GLEMeasureBox measure;
	measure.measureStart();
	// If there are no labels, use g_set_bounds to initialize "measure"
	switch (ax->type) {
		case GLE_AXIS_X:
		case GLE_AXIS_X0:
			g_set_bounds(ox+ax->shift, oy-real_llen);
			break;
		case GLE_AXIS_Y:
		case GLE_AXIS_Y0:
			g_set_bounds(ox-real_llen, oy+ax->shift);
			break;
		case GLE_AXIS_X2:
		case GLE_AXIS_T:
			g_set_bounds(ox+ax->shift, oy+real_llen);
			break;
		case GLE_AXIS_Y2:
			g_set_bounds(ox+real_llen, oy+ax->shift);
			break;
	}
	if (!ax->label_off && drawrest) {
		int nb_names = ax->getNbNamedPlaces();
		for (i = 0; i < nb_names; i++) {
			g_measure(ax->names[i],&bl,&br,&bu,&bd);
			if (bu > maxh) maxh = bu;
			if (-bd > maxd) maxd = -bd;
		}
		int place_cnt = 0;
		double angle = ax->getLabelAngle();
		for (i = 0; i < nb_names; i++) {
			fi = ax->places[i];
			const string& name = ax->names[i];
			if (!ax->isNoPlaceLogOrReg(fi, &place_cnt, dticks) && name != "") {
				fi = fnx(fi);
				if (ax->log) fi = fnlx(ax->places[i]);
				g_measure(name, &bl, &br, &bu, &bd);
				switch (ax->type) {
					case GLE_AXIS_X:
					case GLE_AXIS_X0:
						if (angle == 0.0) {
							g_move(ox+fi+ax->shift, oy-llen-maxh);
							g_jtext(JUST_CENTRE);
						} else {
							g_gsave();
							g_move(ox+fi+ax->shift, oy-llen);
							g_rotate(angle);
							if (angle == 90.0) g_jtext(JUST_RC);
							else if (angle > 0.0) g_jtext(JUST_TR);
							else g_jtext(JUST_TL);
							g_grestore();
						}
						break;
					case GLE_AXIS_Y:
					case GLE_AXIS_Y0:
						if (angle == 0.0) {
							g_move(ox-llen, oy+fi+ax->shift);
							g_jtext(JUST_RC);
						} else {
							g_gsave();
							if (angle == 90.0) {
								g_move(ox-llen-maxd, oy+fi+ax->shift);
								g_rotate(90);
								g_jtext(JUST_CENTRE);
							} else {
								g_move(ox-llen, oy+fi+ax->shift);
								g_rotate(angle);
								if (angle > 0.0) g_jtext(JUST_BR);
								else g_jtext(JUST_TR);
							}
							g_grestore();
						}
						break;
					case GLE_AXIS_X2:
						if (angle == 0.0) {
							g_move(ox+fi+ax->shift, oy+llen+maxd);
							g_jtext(JUST_CENTRE);
						} else {
							g_gsave();
							g_move(ox+fi+ax->shift, oy+llen);
							g_rotate(angle);
							if (angle == 90.0) g_jtext(JUST_LC);
							else if (angle > 0.0) g_jtext(JUST_BL);
							else g_jtext(JUST_BR);
							g_grestore();
						}
						break;
					case GLE_AXIS_Y2:
						g_move(ox+llen, oy+fi+ax->shift);
						g_jtext(JUST_LC);
						break;
				}
			}
		}
	}
	measure.measureEnd();
	// g_box_stroke(measure.getX1(), measure.getY1(), measure.getX2(), measure.getY2(), false);

/*---------------------------------- Now the axis title. */
	if (ax->title_off || ax->title == "" || !drawrest) {
		g_update_bounds_box(box);
		return;
	}
	if (g_get_compatibility() <= GLE_COMPAT_35) {
		draw_axis_titles_v35(ax, h, ox, oy, dticks, llen);
	} else {
		draw_axis_titles(ax, h, ox, oy, measure);
	}
	g_update_bounds_box(box);
}

void draw_axis_titles(axis_struct *ax, double h, double ox, double oy, GLEMeasureBox measure) {
	double bl,br,bu,bd;
	g_gsave();
	double th = h * g_get_fconst(GLEC_ATITLESCALE);
	if (ax->title_scale != 0) th = th*ax->title_scale;
	if (ax->title_hei != 0) th = ax->title_hei;
	g_set_color(ax->title_color);
	g_set_font(ax->title_font);
	g_set_hei(th);
	double tdist = ax->title_dist;
	if (tdist == 0) tdist = h * g_get_fconst(GLEC_ATITLEDIST);
	string title_str = ax->title;
	if (g_get_tex_labels()) {
		title_str.insert(0, "\\tex{");
		title_str.append("}");
	}
	g_measure(title_str,&bl,&br,&bu,&bd);
	switch (ax->type) {
		case GLE_AXIS_X:
		case GLE_AXIS_X0:
			g_move(ox+ax->length/2, measure.getY1() - tdist);
			g_jtext(JUST_TC);
			break;
		case GLE_AXIS_Y:
		case GLE_AXIS_Y0:
			g_move(measure.getX1() - tdist,oy + ax->length/2);
			g_rotate(90.0);
			if (ax->isAlignBase()) g_jtext(JUST_CENTER);
			else g_jtext(JUST_BC);
			break;
		case GLE_AXIS_X2:
		case GLE_AXIS_T:
			g_move(ox+ax->length/2, measure.getY2() + tdist);
			// *should* be center - otherwise titles of different graphs don't align!
			if (ax->isAlignBase()) g_jtext(JUST_CENTER);
			else g_jtext(JUST_BC);
			break;
		case GLE_AXIS_Y2:
			g_move(measure.getX2() + tdist,oy + ax->length/2);
			if (ax->title_rot) {
				g_rotate(-90.0);
				if (ax->isAlignBase()) g_jtext(JUST_CENTER);
				else g_jtext(JUST_BC);
			} else {
				g_rotate(90.0);
				g_jtext(JUST_TC);
			}
			break;
	}
	g_grestore();
}

void draw_axis_titles_v35(axis_struct *ax, double h, double ox, double oy, double dticks, double llen) {
	double bl,br,bu,bd,x,y;
	double ty = 0;
	double tx = ox-.3*h;
	double maxd = 0;
	if (ax->type == GLE_AXIS_Y2) tx = ox + .3*h;
	if (ax->type == GLE_AXIS_X  || ax->type == GLE_AXIS_X0) ty = oy-llen-.3*h;
	if (ax->type == GLE_AXIS_X2 || ax->type == GLE_AXIS_T) ty = oy+llen;
	if (!ax->label_off) {
		int nb_names = ax->getNbNamedPlaces();
		for (int i = 0; i < nb_names; i++) {
			g_measure(ax->names[i],&bl,&br,&bu,&bd);
			if (bd>maxd) maxd = bd;
		}
		int place_cnt = 0;
		for (int i = 0; i < nb_names; i++) {
			double fi = ax->places[i];
			const string& name = ax->names[i];
			if (!ax->isNoPlaceLogOrReg(fi, &place_cnt, dticks) && name != "") {
				fi = fnx(fi);
				if (ax->log) fi = fnlx(ax->places[i]);
				g_measure(name, &bl, &br, &bu, &bd);
				switch (ax->type) {
					case GLE_AXIS_X:
					case GLE_AXIS_X0:
						y = oy-llen-bu+bd-.3*h;
						if (y<ty) ty = y;
						break;
					case GLE_AXIS_Y:
					case GLE_AXIS_Y0:
						x = ox-br+bl-llen-.3*h;
						if (x<tx) tx = x;
						break;
					case GLE_AXIS_X2:
					case GLE_AXIS_T:
						y = oy+llen+maxd+bu;
						if (y>ty) ty = y;
						break;
					case GLE_AXIS_Y2:
						x = ox+br-bl+llen+.3*h;
						if (x>tx) tx = x;
						break;
				}
				if (bd>maxd) maxd = bd;
			}
		}
	}
	g_gsave();
	double th = h * 1.3;
	if (ax->title_scale!=0) th = th*ax->title_scale;
	if (ax->title_hei!=0) th = ax->title_hei;
	g_set_color(ax->title_color);
	g_set_font(ax->title_font);
	g_set_hei(th);
	g_measure(ax->title,&bl,&br,&bu,&bd);
	switch (ax->type) {
		case GLE_AXIS_X:
		case GLE_AXIS_X0:
			g_move(ox+ax->length/2,ty - ax->title_dist);
			g_jtext(JUST_TC);
			break;
		case GLE_AXIS_Y:
		case GLE_AXIS_Y0:
			g_move(tx - ax->title_dist,oy + ax->length/2);
			g_rotate(90.0);
			g_jtext(JUST_BC);
			g_rotate(-90.0);
			break;
		case GLE_AXIS_X2:
		case GLE_AXIS_T:
			g_move(ox+ax->length/2,ty + ax->title_dist);
			g_jtext(JUST_BC);
			break;
		case GLE_AXIS_Y2: /*  y2axis   */
			g_move(tx + ax->title_dist,oy + ax->length/2);
			if (ax->title_rot) {
				g_rotate(-90.0);
				g_jtext(JUST_BC);
				g_rotate(90.0);
			} else {
				g_rotate(90.0);
				g_jtext(JUST_TC);
				g_rotate(-90.0);
			}
			break;
	}
	g_grestore();
}

void nice_ticks(double *dticks, double *gmin, double *gmax, double *t1, double *tn, int minset, int maxset) {
	double delta,st,expnt,n;
	int ni;
	delta = *gmax-*gmin;
	if (delta == 0) {
		gprint("Axis range error min=%g max=%g \n",*gmin,*gmax);
		*gmax = *gmin+10;
		delta = 10;
	}
	st = delta/10;
	expnt = floor(log10(st));
	n = st/pow(10.0,expnt);
	if (n>5)
		ni = 10;
	else if (n>2)
		ni = 5;
	else if (n>1)
		ni = 2;
	else
		ni = 1;
	if (*dticks==0) *dticks = ni * pow(10.0,expnt);
	if (*gmin - (delta/1000) <=  floor( *gmin/ *dticks) * *dticks)
		*t1 = *gmin;
	else
		*t1 = (floor(*gmin/ *dticks) * *dticks ) + *dticks;

	*tn = *gmax;
	if (  (floor(.000001+ *gmax / *dticks) * *dticks) <
	  (*gmax - (delta/1000) ) )
		*tn = floor(.00001 + *gmax/ *dticks ) * *dticks;
}

double chop(double f) {
	return (double) (int) f;
}

void start_subtick(double *tick1, double gmin, double dticks) {
	if (gmin == chop(gmin/dticks) * dticks)
		*tick1 = gmin;
	else
		*tick1 = chop(gmin/dticks) * dticks + dticks;
}

void numtrime(char *o,char *s) {
	char *e,*f;

	strcpy(o,s);

	e = strchr(s,'e');
	if (e==NULL) return;

	e--;
	for (; *e=='0'; e--) ;

	f = strchr(s,'e');
	strcpy(e+1,f);
	strcpy(o,s);
}

void numtrim(char **d,char *s,double dticks) {
	char *o,*nonzero=0;

	if (*d==0)  *d = (char*) myallocz(20);
	o = *d;
	nonzero = 0;
	if (strchr(s,'e')!=NULL) {
		numtrime(o,s);
		return;
	}
	while (*s==' ' && *s!=0) s++;
	while (*s!=0) {
		*(o++) = *(s++);
		if (*s=='.') {
			nonzero = o-1;
			if (dticks!=floor(dticks)) nonzero = o+1;
			while (*s!=0) {
				*(o++) = *(s++);
				if ((*s!='0') && (*s!=0))
					if (o>nonzero) nonzero = o;
			}
		}
	}
	*(o++) = 0;
	if (nonzero!=NULL) *(nonzero+1) = 0;
}

void nice_log_ticks(double *start, double *last, double gmin, double gmax) throw (ParserError) {
	if (gmin <= 0 || gmax <= 0) {
		stringstream err;
		err << "illegal range for log axis: min = ";
		err << gmin << " max = " << gmax;
		g_throw_parser_error(err.str());
	}
	/* range in powers of 10 */
	/* add one extra power on both sides! */
	*start = floor(log10(gmin)-1e-6);
	if (equals_rel(gmin, pow(10.0, *start+1))) {
		(*start)++;
	}
	*last =  ceil(log10(gmax)+1e-6);
	if (equals_rel(gmax, pow(10.0, *last-1))) {
		(*last)--;
	}
}

double fnloglen(double v,axis_struct *ax) {
	if (ax->negate) {
		v = ax->max - (v- ax->min);
	}
	return ((v-log10(ax->min))/(log10(ax->max)-log10(ax->min))) * ax->length;
}

double fnlogx(double v,axis_struct *ax) {
	if (ax->negate) {
		v = ax->max - (v- ax->min);
	}
	return fnloglen(log10(v),ax);
}

int fnbig(double v) {
	if (fabs(log10(v)-floor(log10(v)+.5)) < .001) return true;
	else return false;
}

void print_axis(axis_struct *ax) {
	cout << "Axis ticks off: " << ax->ticks_off << "\tAxis ticks length: "
	     << ax->ticks_length << "\tAxis ticks color: " << ax->ticks_color << endl;
	cout << "Axis subticks off: " << ax->subticks_off << "\tAxis subticks length: "
	     << ax->subticks_length << "\tAxis subticks color: "  << ax->subticks_color << endl;
}

bool axis_is_max(int axis) {
	return axis == GLE_AXIS_X2 || axis == GLE_AXIS_Y2;
}

bool axis_ticks_neg(int axis) {
	return axis == GLE_AXIS_X2 || axis == GLE_AXIS_Y2;
}

bool axis_horizontal(int axis) {
	return axis == GLE_AXIS_X0 || axis == GLE_AXIS_X || axis == GLE_AXIS_X2 || axis == GLE_AXIS_T;
}

int axis_get_orth(int axis, int which) {
	if (axis_horizontal(axis)) {
		switch (which) {
			case 0:  return GLE_AXIS_Y0;
			case 1:  return GLE_AXIS_Y;
			default: return GLE_AXIS_Y2;
		}
	} else {
		switch (which) {
			case 0:  return GLE_AXIS_X0;
			case 1:  return GLE_AXIS_X;
			default: return GLE_AXIS_X2;
		}
	}
}

int axis_type(char *s) {
	if (str_ni_equals(s,"X0",2)) return GLE_AXIS_X0;
	if (str_ni_equals(s,"Y0",2)) return GLE_AXIS_Y0;
	if (str_ni_equals(s,"X2",2)) return GLE_AXIS_X2;
	if (str_ni_equals(s,"Y2",2)) return GLE_AXIS_Y2;
	if (str_ni_equals(s,"X",1))  return GLE_AXIS_X;
	if (str_ni_equals(s,"Y",1))  return GLE_AXIS_Y;
	return GLE_AXIS_ALL;
}

axis_struct::axis_struct() : format() {
	init(0);
}

axis_struct::~axis_struct() {
}

void axis_struct::init(int i) {
	int c;
	format = ""; title = "";
	clearNoTicks();
	names.clear(); places.clear(); noplaces.clear();
	base = 0.0; length = 0.0; shift = 0.0;
	label_font = 0; label_hei = 0.0; label_scale = 0.0; label_dist = 0.0;
	log = 0; min = 0.0; max = 0.0; minset = 0; maxset = 0; nofirst = 0; nolast = 0;
	nticks = 0; nsubticks = 0; dticks = 0.0; dsubticks = 0.0;
	ticks_length = 0.0; ticks_scale = 0.0; ticks_lstyle[0] = 0;
	subticks_length = 0.0; subticks_scale = 0.0; subticks_lstyle[0] = 0;
	label_lstyle[0] = 0;
	off = 0;
	label_off = (i != GLE_AXIS_X && i != GLE_AXIS_Y);
	side_off = 0; ticks_off = 0; subticks_off = 0;
	side_lstyle[0] = 0;
	title_font = 0; title_dist = 0.0; title_hei = 0.0; title_scale = 0.0;
	title_rot = 0; title_off = 0; title_color = 0;
	negate = 0;
	names_ds = -1;
	label_angle = 0.0;
	g_get_color(&c);
	color = c; side_color = c; ticks_color = c;
	label_color = c; subticks_color = c;
	side_lwidth = -1; ticks_lwidth = -1; subticks_lwidth = -1; label_lwidth = -1;
	lgset = GLE_AXIS_LOG_DEFAULT;
	has_ftick = false; ftick = 0.0;
	has_offset = false; offset = 0.0;
	ticks_both = false; grid = false; gridtop = false;
	type = i;
	if (g_get_compatibility() <= GLE_COMPAT_35) {
		alignBase = false;
	} else {
		if (i == GLE_AXIS_Y || i == GLE_AXIS_Y0 || i == GLE_AXIS_Y2) alignBase = false;
		else alignBase = true;
	}
}

string* axis_struct::getNamePtr(int i) {
	while ((int)names.size() <= i) names.push_back(string());
	return &names[i];
}

void axis_struct::setName(int i, const char* name) {
	while ((int)names.size() <= i) names.push_back(string());
	names[i] = name;
}

void axis_struct::setPlace(int i, double place) {
	while ((int)places.size() <= i) places.push_back(0.0);
	places[i] = place;
}

void axis_struct::insertNoTick(double pos, vector<double>& vec) {
	vector<double>::size_type at = 0;
	while (at < vec.size() && vec[at] < pos) {
		at++;
	}
	if (at == vec.size()) vec.push_back(pos);
	else vec.insert(vec.begin()+at, pos);
}

bool axis_struct::isNoPlaceLogOrReg(double pos, int* place_cnt, double delta) {
	if (log) return axis_is_pos_perc(pos, place_cnt, 0.001, noplaces);
	else return axis_is_pos(pos, place_cnt, delta, noplaces);
}

void axis_struct::addNoTick(double pos) {
	addNoTick1(pos);
	addNoTick2(pos);
}

void axis_struct::insertNoTick(double pos) {
	insertNoTick1(pos);
	insertNoTick2(pos);
}

void axis_struct::insertNoTickOrLabel(double pos) {
	insertNoTick1(pos);
	insertNoTick2(pos);
	insertNoPlace(pos);
}

void axis_struct::clearNoTicks() {
	noticks1.clear();
	noticks2.clear();
}

void axis_struct::printNoTicks() {
	cout << "Noticks1:";
	for (vector<double>::size_type i = 0; i < noticks1.size(); i++) {
		cout << " " << noticks1[i];
	}
	cout << endl;
	cout << "Noticks2:";
	for (vector<double>::size_type i = 0; i < noticks2.size(); i++) {
		cout << " " << noticks2[i];
	}
	cout << endl;
	cout << "NoPlaces:";
	for (vector<double>::size_type i = 0; i < noplaces.size(); i++) {
		cout << " " << noplaces[i];
	}
	cout << endl;
}

int axis_struct::getNbNamedPlaces() {
	int nb = getNbNames();
	if (getNbPlaces() < nb) nb = getNbPlaces();
	return nb;
}

void axis_struct::getLabelsFromDataSet(int ds) {
	int crpos = 0;
	int npnts = dp[ds]->np;
	double* xt = dp[ds]->xv;
	if (npnts <= 0) return;
	double min_val = xt[0];
	double max_val = xt[npnts-1];
	double half_spc = (max_val - min_val)/npnts/2.0;
	min_val -= half_spc;
	max_val += half_spc;
	vector<string>* yv_str = dp[ds]->yv_str;
	if (yv_str == NULL) return;
	for (int i = 0; i < getNbPlaces(); i++) {
		double fi = places[i];
		if (fi >= min_val && fi <= max_val) {
			// find last position with x-value smaller than fi
			while (crpos < npnts && xt[crpos] < fi) crpos++;
				if (crpos < npnts && crpos >= 0) {
				if (crpos > 0) crpos--;
				int sel = crpos;
				double dist = fabs(xt[crpos] - fi);
				if (crpos+1 < npnts) {
					if (fabs(xt[crpos+1] - fi) < dist) sel = crpos+1;
				}
				if (crpos-1 >= 0) {
					if (fabs(xt[crpos-1] - fi) < dist) sel = crpos-1;
				}
				if (sel >= 0 && sel < (int)yv_str->size()) *getNamePtr(i) = (*yv_str)[sel];
			}
		}
	}
}
