/*************************************************************************
 *                                                                       *
 * GLE - Graphics Layout Engine <http://www.gle-graphics.org/>           *
 *                                                                       *
 * BSD License                                                           *
 *                                                                       *
 * Copyright (c).                                                        *
 * All rights reserved.                                                  *
 *                                                                       *
 * Redistribution and use in source and binary forms, with or without    *
 * modification, are permitted provided that the following conditions    *
 * are met:                                                              *
 *                                                                       *
 *  * Redistributions of source code must retain the above copyright     *
 *  notice, this list of conditions and the following disclaimer.        *
 *                                                                       *
 *  * Redistributions in binary form must reproduce the above            *
 *  copyright notice, this list of conditions and the following          *
 *  disclaimer in the documentation and/or other materials provided      *
 *  with the distribution.                                               *
 *                                                                       *
 *  * Neither the name of GLE nor the names of its contributors may      *
 *  be used to endorse or promote products derived from this software    *
 *  without specific prior written permission.                           *
 *                                                                       *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR *
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT  *
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, *
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT      *
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, *
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY *
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT   *
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE *
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.  *
 *                                                                       *
 *************************************************************************/

#include "basicconf.h"
#include "parseafm.h"

int debugit=0;
#define dbg if (debugit)
void chrcat(char *s,char c);
int str_ascii(char *s) ;
void writefmt(void);
/*-----------------------------------------------------------------------*/
/* Globals */
FontInfo *fi;
FLAGS myflags = P_ALL;

/*-----------------------------------------------------------------------*/

char fmtname[80];

vector<string> accents;

void init_accents() {
	accents.push_back("caron");
	accents.push_back("cedilla");
	accents.push_back("acute");
	accents.push_back("dieresis");
	accents.push_back("circumflex");
	accents.push_back("grave");
	accents.push_back("ring");
	accents.push_back("tilde");
}

int find_accent(const string& accent) {
	for (vector<string>::size_type i = 0; i < accents.size(); i++) {
		if (accent == accents[i]) {
			return i;
		}
	}
	return -1;
}

int main( int argc, char *argv[])
{
  char *s;
	char filename[80];
	FILE *fp;

	myflags = P_ALL;

	init_accents();

	strcpy(filename,*(++argv));
	s = strchr(filename,'.');
	if (s!=NULL) *s=0;
	strcpy(fmtname,filename);
	if (strchr(filename,':')!=NULL) strcpy(fmtname,strchr(filename,':')+1);
	strcat(fmtname,".fmt");
	strcat(filename,".afm");
	printf("[%s]==>[%s]\n",filename,fmtname);

	if (!filename[0]) {
	  printf ("*** ERROR: can't open. filename is missing.\n");
	  return 0;
	}
	//printf("Reading ADOBE font metric file {%s} \n ",filename);
	//
	fp = fopen(filename,READ_BIN);
	if (fp == NULL) {
		printf ("*** ERROR: can't find: %s\n", filename );
		perror("");
		exit(1);
	}

	switch (parseFile(fp, &fi, myflags)) {
	case parseError:
	    printf("*** ERROR: problem in parsing the AFM File.\n");
	    exit(1);
	case ok:
	    fclose(fp);
	    writefmt();
	    break;
	case earlyEOF:
	    printf("The AFM File ended prematurely.\n");
	    exit(1);
	    break;
	default:
	    break;
    }
    //printf("Successful completion. OK!\n");
  exit(0);
}

/*-----------------------------------------------------------------------*/
struct xypair {float x,y;} ;
struct allchar {char lig[4],lrep[4],kern[35]; struct xypair kxy[35];
} *ap[256];
/* *ap[256];  */
struct char_data { float wx,wy,x1,y1,x2,y2; } cc ;
struct char_datax { char *kern, *lig, *krep, *lrep; };
struct font_table {
	char *name;
	char *full_name;
	char *file_metric;
	char *file_vector;
	char *file_bitmap;
	struct char_data ( *chr)[256];
	int encoding;           /* 1 = postscript text , 2=postscrip symbol, */
				/* 3=TEX text, 4=TEX symbol, 5=TEX extensions */
	float space,space_stretch,space_shrink;
	float scale,slant,uposition,uthickness;
	float fx1,fy1,fx2,fy2,caphei,xhei,descender,ascender;
} fnt;
FILE *fmt;
int c1,c2,ll,lk,n;

void writefmt()
{
	int i = 0,zz;
	double fff;
	CharMetricInfo *temp = fi->cmi;
	Ligature *node = temp->ligs;
	PairKernData *pkd = fi->pkd;
	float scl;
	fff = 2;
	/* sss is never used    a.r. */
	/* sss = (char *) ecvt(fff,2,&dec,&sgn); */

	fmt = fopen(fmtname,WRITE_BIN);

	if (fmt == NULL) {
		perror ("Cant open output file ");
		exit(1);
	}
	fnt.encoding = 1;       /* AdobeStandardEncoding */
	//printf("The font name is %s\n", fi->gfi->fontName);
	if (0==strcmp(fi->gfi->encodingScheme,"TEXENCODING")) {
		fnt.encoding = 3;       /* AdobeStandardEncoding */
	}
	if (0==strcmp(fi->gfi->encodingScheme,"AdobeStandardEncoding")) {
		fnt.encoding = 1;       /* AdobeStandardEncoding */
	}
	if (0==strcmp(fi->gfi->encodingScheme,"GLEMARK")) {
		fnt.encoding = 5;
	}
	//printf("Encoding scheme {%s} %d \n",fi->gfi->encodingScheme,fnt.encoding);
	fnt.scale = 1000;
	scl = .001;
	fnt.slant = fi->gfi->italicAngle;
	fnt.uposition = fi->gfi->underlinePosition*scl;
	fnt.uthickness =  fi->gfi->underlineThickness*scl;
	fnt.fx1 = fi->gfi->fontBBox.llx*scl;
	fnt.fy1 = fi->gfi->fontBBox.lly*scl;
	fnt.fx2 = fi->gfi->fontBBox.urx*scl;
	fnt.fy2 = fi->gfi->fontBBox.ury*scl;
	fnt.caphei = fi->gfi->capHeight*scl;
	fnt.xhei = fi->gfi->xHeight*scl;
	fnt.descender = fi->gfi->descender*scl;
	fnt.ascender = fi->gfi->ascender*scl;

	temp = fi->cmi;
	zz = 1;
	for (temp=fi->cmi;zz < fi->numOfChars; temp++) {
		zz++;
		if (temp->code==32) {
			fnt.space = temp->wx*scl;
			fnt.space_stretch = fnt.space*.5;
			fnt.space_shrink = fnt.space*.3;
			break;
		}
	}

	temp = fi->cmi;
	fwrite(&fnt.encoding,4,16,fmt);

	zz = 0;
	for (i = 0; i < 256 ; ++i) {
	  if (i<temp->code || zz >= fi->numOfChars ) {
		cc.wx = 0;
		cc.wy = 0;
		cc.x1 = 0;
		cc.x2 = 0;
		cc.y1 = 0;
		cc.y2 = 0;
	  } else {
		cc.wx = temp->wx*scl;
		cc.wy = temp->wy*scl;
		cc.x1 = temp->charBBox.llx*scl;
		cc.x2 = temp->charBBox.urx*scl;
		cc.y1 = temp->charBBox.lly*scl;
		cc.y2 = temp->charBBox.ury*scl;
		  for (node=temp->ligs;node !=NULL; node = node -> next) {
		if (ap[i]==NULL) ap[i] = (struct allchar*) calloc(1,sizeof(*ap[i]));
		if (ap[i]==NULL) printf("memory allocation error \n");
			chrcat(ap[i]->lig,str_ascii(node->succ));
			chrcat(ap[i]->lrep,str_ascii(node->lig));
		  }
		zz++;
		temp++;
	  }
	  fwrite(&cc,sizeof(cc),1,fmt);
	} /* for */
	dbg printf("got char %d %d\n",i,zz);

	dbg printf("Here comes the pair kerning data ...\n");
	if (fi->numOfPairs != 0) {
	  for (i = 0; i < fi->numOfPairs; ++i) {
		c1 = str_ascii(pkd[i].name1);
		c2 = str_ascii(pkd[i].name2);
		if (ap[c1]==NULL) ap[c1] = (struct allchar*) calloc(1,sizeof(*ap[c1]));
		if (ap[c1]==NULL) printf("memory allocation error \n");
		chrcat(ap[c1]->kern,c2);
		n = strlen(ap[c1]->kern)-1;
		ap[c1]->kxy[n].x = pkd[i].xamt*scl;
		ap[c1]->kxy[n].y = pkd[i].yamt*scl;
	  }
	}

	for (i=1;i<256;i++) {
		if (ap[i]==NULL) continue;
		ll = strlen(ap[i]->lig);
		lk = strlen(ap[i]->kern);
		if (ll>0 || lk>0) {
			fputc(i,fmt);
			fputc(ll,fmt);
			fwrite(ap[i]->lig,1,ll,fmt);
			fwrite(ap[i]->lrep,1,ll,fmt);
			fputc(lk,fmt);
			fwrite(ap[i]->kern,1,lk,fmt);
			fwrite(ap[i]->kxy,8,lk,fmt);
		}
	}
	fputc(0,fmt);

	/* Also do the composites to support accents in GLE (new 070506) */
	for (i = 0; i < fi->numOfComps; i++) {
		CompCharData *ccd = &(fi->ccd[i]);
		string name = ccd->ccName;
		if (name.length() > 1) {
			int char_code = name[0];
			string accent_code = name.substr(1);
			int accent_idx = find_accent(accent_code);
			if (accent_idx != -1) {
				if (ccd->numOfPieces != 2) {
					cout << ">>> incorrect number of pieces: '" << name << "': " << ccd->numOfPieces << endl;
				} else {
					fwrite(&char_code, sizeof(int), 1, fmt);
					fwrite(&accent_idx, sizeof(int), 1, fmt);
					for (int j = 0; j < ccd->numOfPieces; j++) {
						Pcc* piece = &ccd->pieces[j];
						int piece_code = str_ascii(piece->pccName);
						if (piece_code == 0) {
							cout << ">>> can't find piece: '" << piece->pccName << "'" << endl;
						}
						fwrite(&piece_code, sizeof(int), 1, fmt);
						double dx = piece->deltax * scl;
						double dy = piece->deltay * scl;
						fwrite(&dx, sizeof(double), 1, fmt);
						fwrite(&dy, sizeof(double), 1, fmt);
					}
				}
			} else {
				cout << ">>> unknown accent: '" << accent_code << "'" << endl;
			}
		} else {
			cout << ">>> composite too short: '" << name << "'" << endl;
		}
	}
	i = 0;
	fwrite(&i, sizeof(int), 1, fmt);

	fclose(fmt);
}
void chrcat(char *s,char c)
{
	int i;
	i = strlen(s);
	*(s+i) = c;
}
int str_ascii(char *s)
{
	int i;
	CharMetricInfo *temp = fi->cmi;

	for (i=0;i<fi->numOfChars; ++i) {
		if (strcmp(temp->name,s)==0) {
			return temp->code;
		}
		temp++;
	}
	printf("Failed to find char, fatal error ********* {%s} \n",s);
	return 0;
}
